@@ -288,9 +288,14 @@ render_basic (PangoFont *font, gunichar wc,
   pango_glyph_string_set_size (glyphs, *n_glyphs + 1);
   if (index)
     set_glyph (font, glyphs, *n_glyphs, cluster_offset, index);
-  else 
-    set_glyph (font, glyphs, *n_glyphs, cluster_offset,
-	       get_unknown_glyph (font, wc));
+  else
+    {
+      if (pango_is_zero_width (wc))
+	set_glyph (font, glyphs, *n_glyphs, cluster_offset, 0);
+      else
+	set_glyph (font, glyphs, *n_glyphs, cluster_offset,
+		   get_unknown_glyph (font, wc));
+    }
   (*n_glyphs)++;
 }
 
@@ -307,7 +312,7 @@ hangul_engine_shape (PangoEngineShape *engine,
   const char *p, *start;
 
   gunichar jamos_static[8];
-  guint max_jamos = G_N_ELEMENTS (jamos_static);
+  gint max_jamos = G_N_ELEMENTS (jamos_static);
   gunichar *jamos = jamos_static;
   int n_jamos;
 
