@@ -659,11 +659,7 @@ itemize_state_init (ItemizeState      *state,
   /* First, apply the bidirectional algorithm to break
    * the text into directional runs.
    */
-  text_ucs4 = g_utf8_to_ucs4_fast (text + start_index, length, &n_chars);
-  state->embedding_levels = g_new (guint8, n_chars);
-  pango_log2vis_get_embedding_levels (text_ucs4, n_chars, &base_dir,
-                                      state->embedding_levels);
-  g_free (text_ucs4);
+  state->embedding_levels = pango_log2vis_get_embedding_levels (text + start_index, length, &base_dir);
   
   state->embedding_end_offset = 0;
   state->embedding_end = text + start_index;
@@ -1158,7 +1154,7 @@ itemize_state_finish (ItemizeState *state)
  *             This must be >= 0.
  * @base_dir:  base direction to use for bidirectional processing
  * @attrs:     the set of attributes that apply to @text.
- * @cached_iter:      Cached attribute iterator, or NULL
+ * @cached_iter:      Cached attribute iterator, or %NULL
  *
  * Like pango_itemize(), but the base direction to use when
  * computing bidirectional levels (see pango_context_set_base_dir ()),
@@ -1231,7 +1227,7 @@ itemize_with_font (PangoContext               *context,
  *             after @start_index.
  *             This must be >= 0.
  * @attrs:     the set of attributes that apply to @text.
- * @cached_iter:      Cached attribute iterator, or NULL
+ * @cached_iter:      Cached attribute iterator, or %NULL
  *
  * Breaks a piece of text into segments with consistent
  * directional level and shaping engine. Each byte of @text will
@@ -1353,7 +1349,7 @@ update_metrics_from_items (PangoFontMetrics *metrics,
  *            The process of determining such metrics is slow, so it is best
  *            to always make sure some real language tag will be used.
  * 
- * Get overall metric information for a font particular font
+ * Get overall metric information for a particular font
  * description.  Since the metrics may be substantially different for
  * different scripts, a language tag can be provided to indicate that
  * the metrics should be retrieved that correspond to the script(s)
