@@ -129,19 +129,14 @@ _pango_xft_font_get_mini_font (PangoXftFont *xfont)
   if (!xfont->mini_font)
     {
       Display *display;
-      int screen;
       PangoFontDescription *desc = pango_font_description_new ();
-      PangoContext *context;
       int i;
       int width = 0, height = 0;
       XGlyphInfo extents;
       XftFont *mini_xft;
       int new_size;
       
-      _pango_xft_font_map_get_info (fcfont->fontmap, &display, &screen);
-
-      context = pango_xft_get_context (display, screen);
-      pango_context_set_language (context, pango_language_from_string ("en"));
+      _pango_xft_font_map_get_info (fcfont->fontmap, &display, NULL);
 
       pango_font_description_set_family_static (desc, "monospace");
 
@@ -152,10 +147,8 @@ _pango_xft_font_get_mini_font (PangoXftFont *xfont)
       else
 	pango_font_description_set_size (desc, new_size);
 
-      xfont->mini_font = pango_font_map_load_font (fcfont->fontmap, context, desc);
-
+      xfont->mini_font = pango_font_map_load_font (fcfont->fontmap, NULL, desc);
       pango_font_description_free (desc);
-      g_object_unref (context);
       
       mini_xft = xft_font_get_font (xfont->mini_font);
       
@@ -365,7 +358,7 @@ load_fallback_font (PangoXftFont *xfont)
   _pango_xft_font_map_get_info (fcfont->fontmap, &display, &screen);
   
   size_is_absolute = pango_font_description_get_size_is_absolute (fcfont->description);
-  size = (double)pango_font_description_get_size (fcfont->description) / PANGO_SCALE;
+  size = pango_font_description_get_size (fcfont->description) / PANGO_SCALE;
       
   xft_font = XftFontOpen (display,  screen,
                           FC_FAMILY, FcTypeString, "sans",
@@ -475,7 +468,7 @@ pango_xft_font_real_shutdown (PangoFcFont *fcfont)
  *
  * Returns the XftFont of a font.
  *
- * Return value: the XftFont associated to @font.
+ * Returns: the XftFont associated to @font.
  **/
 XftFont *
 pango_xft_font_get_font (PangoFont *font)
@@ -491,7 +484,7 @@ pango_xft_font_get_font (PangoFont *font)
  *
  * Returns the X display of the XftFont of a font.
  *
- * Return value: the X display of the XftFont associated to @font.
+ * Returns: the X display of the XftFont associated to @font.
  **/
 Display *
 pango_xft_font_get_display (PangoFont *font)
@@ -538,7 +531,7 @@ pango_xft_font_get_unknown_glyph (PangoFont *font,
  *
  * Use pango_fc_font_lock_face() instead.
  *
- * Return value: the FreeType FT_Face associated with @font.
+ * Returns: the FreeType FT_Face associated with @font.
  *
  * Since: 1.2
  **/
